package gui;

import java.awt.*;
import java.awt.geom.Ellipse2D;

import javax.swing.JPanel;

import model.Tape;

/**
 * A panel which shows the tape.
 */
public class TapePanel extends JPanel {

    /**
     * The size (width and height) of one cell.
     */
    public static final int CELL_SIZE = 50;

    /**
     * The size (in pixels) of the margin above and below the tape.
     */
    public static final int MARGIN = 5;

    /**
     * The tape to show in the visualization.
     */
    private Tape<Character> tape;

    /**
     * Creates a tape panel with the given tape.
     *
     * @param tape The tape to show.
     */
    public TapePanel(Tape<Character> tape) {
        this.tape = tape;
    }

    @Override
    public Dimension getPreferredSize() {
        return new Dimension(10 * CELL_SIZE + 1, CELL_SIZE + 2 * MARGIN + 1);
    }

    /**
     * Sets the tape shown by this panel.
     *
     * @param tape The new tape to show.
     */
    public void setTape(Tape<Character> tape) {
        this.tape = tape;
    }

    @Override
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);

        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                RenderingHints.VALUE_TEXT_ANTIALIAS_LCD_HRGB);

        setFont(getFont().deriveFont(CELL_SIZE * 0.6f).deriveFont(Font.BOLD));
        // determine whether the tape content is larger than what is currently
        // shown on screen
        int w = getWidth() / (2 * CELL_SIZE);
        int leftBound;
        int rightBound;
        
        if (-w - tape.getPosition() < -tape.getPosition()) {
            leftBound = -w - tape.getPosition();
        } else {
            leftBound = -tape.getPosition();
        }
        
        if (w > tape.getSize() - tape.getPosition()) {
            rightBound = w;
        } else { 
            rightBound = tape.getSize() - tape.getPosition();
        }
        
        // System.out.println("leftBound: " + leftBound);
        // System.out.println("rightBound: " + rightBound);
        
        // the "dots" to show the position better
        for (int i = leftBound % 10;
                i <= rightBound;
                i += 10) {
            drawDots(g2, i);
        }

        // the cells
        for (int i = leftBound - 1;
                i <= rightBound + 1;
                i++) {
            drawCell(g2, i);
        }

        // decoration for the tape head
        g2.setColor(Color.BLACK);
        g2.drawRect(getWidth() / 2 - CELL_SIZE / 2 - 1,
                MARGIN - 1, CELL_SIZE + 2, getHeight() - 2 * MARGIN + 2);
        g2.drawRect(getWidth() / 2 - CELL_SIZE / 2 - 2,
                MARGIN - 2, CELL_SIZE + 4, getHeight() - 2 * MARGIN + 4);
    }

    /**
     * Draws the <code>i</code><sup>th</sup> cell of the tape.
     *
     * @param g2 The graphics object to draw with.
     * @param i The index of the tape cell. (See the documentation of
     * {@link Tape#peek(int)} for the definition of this index.)
     */
    private void drawCell(Graphics2D g2, int i) {

        // background
        if (i == 0) {
            g2.setColor(new Color(235, 235, 235));
        } else {
            g2.setColor(Color.WHITE);
        }
        g2.fillRect(getWidth() / 2 - CELL_SIZE / 2 + i * CELL_SIZE,
                MARGIN, CELL_SIZE, getHeight() - 2 * MARGIN);

        // border
        g2.setColor(Color.BLACK);
        g2.drawRect(getWidth() / 2 - CELL_SIZE / 2 + i * CELL_SIZE,
                MARGIN, CELL_SIZE, getHeight() - 2 * MARGIN);

        // the symbol
        Character symbol = tape == null ? null : tape.peek(i);

        if (symbol == null) {
            // if it is a blank, draw a small square in gray
            g2.setColor(Color.LIGHT_GRAY);

            g2.drawRect(getWidth() / 2 + i * CELL_SIZE - 5,
                    getHeight() / 2 - 5, 10, 10);
        } else {
            // if it is not a blank, just draw the symbol
            g2.setColor(Color.BLACK);
            FontMetrics fm = g2.getFontMetrics();
            int width = fm.stringWidth(symbol.toString());
            int height = fm.getAscent();
            g2.drawString(symbol.toString(),
                    getWidth() / 2 + i * CELL_SIZE - width / 2,
                    getHeight() / 2 + height / 2);
        }
    }

    /**
     * Draws dots for the the <code>i</code><sup>th</sup> cell of the tape.
     *
     * @param g2 The graphics object to draw with.
     * @param i The index of the tape cell. (See the documentation of
     * {@link Tape#peek(int)} for the definition of this index.)
     */
    private void drawDots(Graphics2D g2, int i) {

        g2.setColor(Color.BLACK);
        g2.fill(new Ellipse2D.Double(
                getWidth() / 2 + i * CELL_SIZE - 4,
                MARGIN - 4,
                8,
                8
        ));
        g2.fill(new Ellipse2D.Double(
                getWidth() / 2 + i * CELL_SIZE - 4,
                getHeight() - MARGIN - 4,
                8,
                8
        ));
    }
}
