package model;

import java.util.ArrayList;

/**
 * This class implements an infinite tape.
 *
 * Objects from type <code>E</code> will be used as tape symbols;
 * <code>null</code> represents a blank.
 *
 * @param <E> The type of symbols on the tape.
 */
public class Tape<E> {

    ArrayList<E> tape;  //create an array list that we use for the tape
    int currentSymbol;  //variable to keep track of the position in the arraylist of the currentSymbol ( under the tape head)

    /**
     * Constructs a new tape, where all elements are initialized to blanks
     * (<code>null</code>).
     */
    public Tape() {
        tape = new ArrayList<E>();  // creates an empty arraylist
        tape.add(null);             // add a blank to the array list
        currentSymbol = 0;    // sets the position in the arraylies of the current symbol to the beginning of the arraylist

    }

    /**
     * Returns the current symbol under the tape head.
     *
     * @return The symbol under the tape read.
     */
    public E readSymbol() {
        return tape.get(currentSymbol); //get the current symbol from arraylist

    }

    /**
     * Writes the given symbol to the current position under the tape head.
     *
     * @param newSymbol The new symbol to write.
     */
    public void writeSymbol(E newSymbol) {
        tape.set(currentSymbol, newSymbol); // sets current symbol to the new symbol and place it in the arraylist on position under the header
    }

    /**
     * Shifts the tape head one place to the left.
     */
    public void shiftLeft() {
        if (currentSymbol == 0) {
            tape.add(0, null);  //adds a blanc in front of the arraylist, current symbol then gets +1 and by shifting one to the left -1; to conclude it stays the same
        } else {
            currentSymbol--;    //currentSymbol minus one "shift one to the left"
        }
    }

    /**
     * Shifts the tape head one place to the right.
     */
    public void shiftRight() {
        if (currentSymbol + 1 == tape.size()) {     // to prevent errors we add null to the end of the array if the "new" posiotion is not defined yet
            tape.add(null);
        }
        currentSymbol++;

    }

    /**
     * Places the given input on the tape. This is called to initialize the tape
     * on the start of a simulation.
     *
     * This method should put the given symbols on the tape from left to right
     * (so the first character of the string should be placed on the leftmost
     * cell, and so on). When the method finishes, the tape head should be
     * placed on the first character of the input.
     *
     * @param input The input symbols to place on the tape.
     */
    public void placeInput(ArrayList<E> input) {
        int tapeHead = currentSymbol;   //Store begin position
        for (int i = 0; i < input.size(); i++) { //iterate through input
            tape.add(currentSymbol, input.get(i));  //add every element of the input to the arraylist
            currentSymbol++;
        }
        currentSymbol = tapeHead;   //Set current postion in the array to before the input was processed
    }

    /**
     * Reads the output that is currently on the tape. In the block currently
     * surrounded by blanks.
     * 
     * The result should be a list of the symbols on the tape, starting from the
     * most left tape symbol, to the most right symbol.
     * <pre>
     * # # a b c # # d e f # #
     *       ^
     * </pre> this method will return <code>"abc def"</code>.
     *
     * @return The output on the tape.
     */
    public ArrayList<E> readCurrentBlockOutput() {
        int displacement = 0;
        ArrayList<E> returnArray = new ArrayList<E>();  //creates a new arraylist that will function as the return arraylist
        if (tape.get(currentSymbol) == null) { //returns blank if a blank is read
            return null;
        } else {
            while (tape.get(currentSymbol) != null) {
                shiftLeft();    //aslong as there is a non-empty symbol to the left go left ( to get to the first symbol of the output);
                displacement--;
            }
            shiftRight();   // shift right once inorder to get to the first symbol of the output
            displacement++;
            while (tape.get(currentSymbol) != null) {
                returnArray.add(tape.get(currentSymbol)); // keep adding symbol to the return array and shifting right untill you find a blank
                shiftRight();
                displacement++;
            }
            // return the tape head to where it was before we got the output from the tape
            while (displacement != 0) {
                shiftLeft();
                displacement--;
            }
            return returnArray;
        }
    }
    
   /**
     * Reads the output that is currently on the tape.
     *
     * The result should be a list of the non-blank symbols on the tape, from
     * left to right. Reading the output stops at blank symbols, so for example
     * on the tape
     * <pre>
     * # # a b c # # d e f # #
     *       ^
     * </pre> this method will return <code>"abc"</code>.
     *
     * @return The output on the tape.
     */
    public ArrayList<Character> readOutput() {
        ArrayList<Character> returnArray = new ArrayList<Character>();  //creates a new arraylist that will function as the return arraylist
        
        for (E c : tape) {
            if (c == null) {
                returnArray.add('#');
            } else {
                returnArray.add((Character) c);
            }
        }
        
        return returnArray;
    }

    /**
     * Returns a symbol that is not currently under the tape head. This should
     * of course not be used in the Turing machine emulator (hence the name
     * <code>peek</code>), but it is used in the visualisation.
     *
     * @param location The location to read from. <code>0</code> stands for the
     * position of the tape head; the elements to the left of it are numbered
	 * <code>-1, -2, -3, ...</code>; the elements to the right of it are
     * numbered <code>1, 2, 3, ...</code>.
     * @return The symbol at the specified location.
     */
    public E peek(int location) {
        int peekSymbol = currentSymbol + location;  // position of the peek symbol in the arraylist

        if ((peekSymbol < 0) || (peekSymbol > tape.size() - 1)) {    //if the position is not defined in the arraylist ( it is a blank then) return a blank
            return null;
        } else {
            return tape.get(peekSymbol); // if it is defined return the good symbol
        }
    }

    /**
     * Returns the absolute position on the tape. This should of course not be
     * used in the Turing machine emulator, but it is used in the visualisation.
     *
     * @return The absolute position on the tape.
     */
    public int getPosition() {
        return currentSymbol;
    }

    /**
     * Resets the tape, that is, ensures that the entire tape is empty.
     *
     * A call to {@link #getPosition()} should return <code>0</code> after this
     * method returns.
     */
    public void reset() {
        tape.clear();   //clear whole array
        tape.add(null); // add a blank to start with
        currentSymbol = 0;  //set header to that blank
    }
    
    /**
     * Returns the size of the tape.
     *
     */
    public int getSize() {
        return tape.size();
    }
}
