package model;

/**
 * A transition is a rule that states how the Turing machine can move from one
 * state to another. Transitions correspond to the "arrows" in graphical
 * representations of Turing machines.
 *
 * A transition has the following elements, that determine how the Turing
 * machine behaves when the transition is taken:
 * <ul>
 * <li>the <b>source state</b>: the state the Turing machine should be in when
 * taking this transition;</li>
 * <li>the <b>target state</b>: the state the Turing machine gets in after
 * having taken this transition;</li>
 * <li>the <b>read symbol</b>: the symbol that needs to be under the tape head
 * when taking this transition;</li>
 * <li>the <b>write symbol</b>: the symbol that is written while taking this
 * transition;</li>
 * <li>the <b>head direction</b>: a value <i>L</i> or <i>R</i> that indicates
 * which direction the tape head should move to.</li>
 * </ul>
 *
 * Transition objects are immutable.
 */
public class Transition {

    /**
     * The source state.
     */
    private final String sourceState;

    /**
     * The target state.
     */
    private final String targetState;

    /**
     * The read symbol.
     */
    private final Character readSymbol;

    /**
     * The write symbol.
     */
    private final Character writeSymbol;

    /**
     * The head direction.
     *
     * TODO [ws] this should probably not be stored as a String, perhaps a
     * boolean or enum would be better
     */
    private final String headDirection;

    /**
     * A string that identifies this transition.
     */
    private final String transitionId;

    /**
     * Creates a new transition.
     *
     * @param sourceState The source state.
     * @param targetState The target state.
     * @param readSymbol The read symbol. (May be <code>"#"</code>, representing
     * a blank.)
     * @param writeSymbol The write symbol. (May be <code>"#"</code>,
     * representing a blank.)
     * @param headDirection The head direction, as a String <code>"L"</code> or
     * <code>"R"</code>.
     */
    public Transition(String sourceState, String targetState,
            char readSymbol, char writeSymbol,
            String headDirection) {

        this.sourceState = sourceState;
        this.readSymbol = readSymbol == '#' ? null : readSymbol;
        this.writeSymbol = writeSymbol == '#' ? null : writeSymbol;
        this.headDirection = headDirection;
        this.targetState = targetState;
        this.transitionId = sourceState + readSymbol + writeSymbol + headDirection + targetState;
    }

    /**
     * Returns the source state of the transition.
     *
     * @return The source state.
     */
    public String getSourceState() {
        return sourceState;
    }

    /**
     * Returns the target state of the transition.
     *
     * @return The target state.
     */
    public String getTargetState() {
        return targetState;
    }

    /**
     * Returns the symbol read from the tape.
     *
     * @return The read symbol.
     */
    public Character getReadTape() {
        return readSymbol;
    }

    /**
     * Returns the symbol written to the tape.
     *
     * @return The write symbol.
     */
    public Character getWriteTape() {
        return writeSymbol;
    }

    /**
     * Returns the head direction.
     *
     * @return The head direction, as a String <code>"L"</code> or
     * <code>"R"</code>.
     */
    public String getHeadDirection() {
        return headDirection;
    }

    /**
     * Returns the ID of this transition.
     *
     * @return The ID.
     */
    public String getTransitionId() {
        return transitionId;
    }

    @Override
    public String toString() {
        return this.sourceState + " "
                + symbolToString(this.readSymbol) + "/"
                + symbolToString(this.writeSymbol) + ","
                + this.headDirection + " "
                + this.targetState;
    }

    /**
     * Helper method for {@link #toString()} that creates a String
     * representation of a tape character.
     *
     * @param c The tape character.
     * @return If <code>c == null</code>, <code>"#"</code> (representing a blank
     * on the tape). Else, a String containing just <code>c</code>.
     */
    private static String symbolToString(Character c) {
        if (c == null) {
            return "#";
        }
        return c.toString();
    }
}
