package model;

import java.util.HashMap;

/**
 * This class implements the behaviour of a Turing machine.
 */
public class TuringMachine {

    Tape tape;
    TuringProgram turingProgram;

    String currentState = "q0";
    int stepCount = 0;

    /**
     * Creates a new Turing machine with an empty tape, that will execute the
     * given Turing program.
     *
     * The Turing machine should start in state "q0".
     *
     * @param program The program to execute.
     */
    public TuringMachine(TuringProgram program) {
        turingProgram = program; //collects from constructor
        this.tape = new Tape();  //creates tape

    }

    /**
     * Returns the tape of the Turing machine.
     *
     * @return The tape.
     */
    public Tape<Character> getTape() {
        return tape;
    }

    /**
     * Returns the name of the current state.
     *
     * @return The current state name.
     */
    public String getCurrentState() {
        return currentState;

    }

    /**
     * Returns the Turing program this machine is executing.
     *
     * @return The Turing program of the machine.
     */
    public TuringProgram getTuringProgram() {
        return turingProgram;
    }

    /**
     * Checks which transition is possible at the moment. (The step is not
     * actually executed, however.)
     *
     * @return The transition that can be done at the moment, or
     * <code>null</code> if no transition can be done anymore. If the Turing
     * machine can perform more than one step at the moment, any of the possible
     * transitions may be returned by this method.
     */
    public Transition activeTransition() {
        /**
         * From the program, we acquire the hashmap of states from. From this
         * map, we collect the current state by searching for is name.
         */
        State sCurrentState = turingProgram.getStates().get(currentState);
        //we collect the map of transitions from the current state
        HashMap<String, Transition> outGoingTrans = sCurrentState.getOutgoingTransitions();

        //we iterate through the map of transitions until we find one which the appropriate condition
        for (Transition returnTrans : outGoingTrans.values()) {
            if (returnTrans.getReadTape() == tape.readSymbol()) {
                return returnTrans;
            }
        }
        //if no transition is found, we simply return null
        return null;
    }

    /**
     * Performs a single transition.
     *
     * If no step could be performed at all, nothing should happen.
     *
     * If the Turing machine can perform more than one step at the moment, any
     * of the possible transitions may be taken.
     */
    public void step() {
        
        //first check whether we can actually do this step
        Transition stepTransition = activeTransition();

        if (activeTransition() != null) {
            //write new symbol, change tape header
            tape.writeSymbol(stepTransition.getWriteTape());
            if (stepTransition.getHeadDirection().equals("L")) {
                tape.shiftLeft();
            } else if (stepTransition.getHeadDirection().equals("R")) {
                tape.shiftRight();
            }
            //change current state to target state
            currentState = stepTransition.getTargetState();
            stepCount++;
        }
    }

    /**
     * Returns the number of steps that have been executed until now.
     *
     * @return The step count.
     */
    public int getStepCount() {
        return stepCount;
    }

    /**
     * Resets the Turing machine.
     *
     * This method restores the Turing machine to the state like it was when it
     * was constructed, that is:
     * <ul>
     * <li>it empties the tape;</li>
     * <li>it resets the current state to "q0";</li>
     * <li>it resets the step count.</li>
     * </ul>
     */
    public void reset() {
        tape.reset();
        currentState = "q0";
        stepCount = 0;
    }
}
